@@ -1,22 +1,16 @@
-{-# LANGUAGE DeriveAnyClass    #-}
-{-# LANGUAGE DeriveFunctor     #-}
-{-# LANGUAGE DeriveGeneric     #-}
-{-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE TypeFamilies      #-}
+{-# LANGUAGE DeriveAnyClass #-}
+{-# LANGUAGE DeriveGeneric  #-}
 
 module A ( E (..), T (..), (~>), TB (..), C (..)
          , L (..), N (..), BBin (..), BTer (..)
          , BUn (..), DfnVar (..)
          , D (..), Program (..)
-         , Mode (..)
+         , Mode (..), awk
          , mapExpr
          , getS, flushD
-         -- * Base functors
-         , EF (..)
          ) where
 
 import           Control.DeepSeq    (NFData)
-import           Control.Recursion  (Base, Corecursive, Recursive)
 import qualified Data.ByteString    as BS
 import qualified Data.IntMap        as IM
 import           Data.List          (foldl')
@@ -26,16 +20,16 @@ import           Data.Text.Encoding (decodeUtf8)
 import qualified Data.Vector        as V
 import           GHC.Generics       (Generic)
 import           Nm
+import           Nm.Map             (NmMap)
+import qualified Nm.Map             as Nm
 import           Prettyprinter      (Doc, Pretty (..), braces, brackets, concatWith, encloseSep, flatAlt, group, hardline, indent, parens, pipe, punctuate, tupled, (<+>))
 import           Regex.Rure         (RurePtr)
 
 infixr 6 <#>
 infixr 6 <##>
 
-(<#>) :: Doc a -> Doc a -> Doc a
+(<#>), (<##>) :: Doc a -> Doc a -> Doc a
 (<#>) x y = x <> hardline <> y
-
-(<##>) :: Doc a -> Doc a -> Doc a
 (<##>) x y = x <> hardline <> hardline <> y
 
 data TB = TyI | TyFloat | TyStr
@@ -49,8 +43,10 @@ tupledByFunky sep = group . encloseSep (flatAlt "⟨ " "⟨") (flatAlt " ⟩" "
 tupledBy :: Doc ann -> [Doc ann] -> Doc ann
 tupledBy sep = group . encloseSep (flatAlt "( " "(") (flatAlt " )" ")") sep
 
-jacTup :: Pretty a => [a] -> Doc ann
-jacTup = tupledBy " . " . fmap pretty
+j'Tup :: Pretty a => [a] -> Doc ann
+j'Tup = tupledBy " . " . fmap pretty
+
+jrec = encloseSep (flatAlt "#{ " "#{") (flatAlt (hardline <+> "}") "}") (flatAlt " ; " "; ")
 
 infixr 0 ~>
 
@@ -60,12 +56,14 @@ infixr 0 ~>
 infixr 0 :$
 
 data T = TyB { tyBuiltin :: TB }
-       | (:$) { tyApp0 :: T, tyApp1 :: T }
-       | TyArr { tyArr0 :: T, tyArr1 :: T }
+       | (:$) { tyApp0, tyApp1 :: T }
+       | TyArr { tyArr0, tyArr1 :: T }
        | TyVar { tyVar :: Nm () }
        | TyTup { tyTups :: [T] }
+       | TyRec { tyres :: NmMap T }
        | Rho { tyRho :: Nm (), tyArms :: IM.IntMap T }
-       deriving (Eq, Ord)
+       | Ρ { tyΡ :: Nm (), tyρs :: NmMap T }
+       deriving Eq
 
 instance Pretty TB where
     pretty TyI = "Integer"; pretty TyStr = "Str"; pretty TyFloat = "Float"
@@ -79,12 +77,14 @@ instance Pretty T where
     pretty (ty:$ty')      = pretty ty <+> pretty ty'
     pretty (TyVar n)      = pretty n
     pretty (TyArr ty ty') = pretty ty <+> "⟶" <+> pretty ty'
-    pretty (TyTup tys)    = jacTup tys
+    pretty (TyTup tys)    = j'Tup tys
+    pretty (TyRec rs)     = jrec ((\(n,t) -> pretty n <> ":" <+> pretty t)<$>Nm.toList rs)
     pretty (Rho n fs)     = braces (pretty n <+> pipe <+> prettyFields (IM.toList fs))
+    pretty (Ρ n fs)       = braces (pretty n <+> pipe <+> prettyFields (Nm.toList fs))
 
 parensp True=parens; parensp False=id
 
-prettyFields :: [(Int, T)] -> Doc ann
+prettyFields :: Pretty a => [(a, T)] -> Doc ann
 prettyFields = mconcat . punctuate "," . fmap g where g (i, t) = pretty i <> ":" <+> pretty t
 
 instance Show T where show=show.pretty
@@ -92,7 +92,7 @@ instance Show T where show=show.pretty
 data BUn = Tally -- length of string field
          | Const
          | Not -- ^ Boolean
-         | At Int | Select Int
+         | At !Int | Select !Int | SelR !(Nm ())
          | IParse | FParse | Parse
          | Floor | Ceiling
          | Some
@@ -108,6 +108,7 @@ instance Pretty BUn where
     pretty Not        = "!"
     pretty (At i)     = "." <> pretty i
     pretty (Select i) = "->" <> pretty i
+    pretty (SelR r)   = "->" <> pretty r
     pretty IParse     = ":i"
     pretty FParse     = ":f"
     pretty Floor      = "floor"
@@ -124,10 +125,10 @@ instance Pretty BUn where
     pretty Last       = "last#"
 
 data BTer = ZipW
-          | Fold | Scan
+          | Fold | Scan | ScanList
           | Substr | Sub1 | Subs
           | Option
-          | Captures | AllCaptures
+          | Captures | AllCaptures | Ixes
           | Bookend
           deriving (Eq)
 
@@ -135,6 +136,7 @@ instance Pretty BTer where
     pretty ZipW        = ","
     pretty Fold        = "|"
     pretty Scan        = "^"
+    pretty ScanList    = "^*"
     pretty Substr      = "substr"
     pretty Option      = "option"
     pretty Captures    = "~*"
@@ -142,6 +144,7 @@ instance Pretty BTer where
     pretty Sub1        = "sub1"
     pretty Subs        = "subs"
     pretty Bookend     = ",,"
+    pretty Ixes        = "ices"
 
 -- builtin
 data BBin = Plus | Times | Div
@@ -158,6 +161,7 @@ data BBin = Plus | Times | Div
           | Match | Sprintf
           | Report
           | Take | Drop
+          | Rein | Nier
           deriving (Eq)
 
 instance Pretty BBin where
@@ -169,11 +173,12 @@ instance Pretty BBin where
     pretty Split = "split"; pretty Splitc = "splitc"; pretty Sprintf = "sprintf"
     pretty Match = "match"; pretty MapMaybe = ":?"; pretty Fold1 = "|>"
     pretty Exp = "**"; pretty DedupOn = "~.*"; pretty Report = "$>"
-    pretty Take = "take#"; pretty Drop = "drop#"
+    pretty Take = "take#"; pretty Drop = "drop#"; pretty Rein = "reintercalate"
+    pretty Nier = "@@"
 
 data DfnVar = X | Y deriving (Eq)
 
-instance Pretty DfnVar where pretty X = "x"; pretty Y = "y"
+instance Pretty DfnVar where pretty X="x"; pretty Y="y"
 
 -- 0-ary
 data N = Ix | Nf | None | Fp | MZ deriving (Eq)
@@ -191,11 +196,10 @@ data E a = Column { eLoc :: a, col :: Int }
          | IParseAllCol { eLoc :: a } -- ^ @$0@, parsed as an integer
          | FParseAllCol { eLoc :: a } -- ^ @$0@, parsed as a float
          | ParseAllCol { eLoc :: a }
-         | EApp { eLoc :: a, eApp0 :: E a, eApp1 :: E a }
-         | Guarded { eLoc :: a, eP :: E a, eGuarded :: E a }
+         | EApp { eLoc :: a, eApp0, eApp1 :: E a }
+         | Guarded { eLoc :: a, eP, eGuarded :: E a }
          | Implicit { eLoc :: a, eImplicit :: E a }
          | Let { eLoc :: a, eBind :: (Nm a, E a), eE :: E a }
-         -- TODO: literals type (make pattern matching easier down the road)
          | Var { eLoc :: a, eVar :: !(Nm a) }
          | F { hole :: !(Nm a) }
          | Lit { eLoc :: a, lit :: !L }
@@ -205,47 +209,16 @@ data E a = Column { eLoc :: a, col :: Int }
          | BB { eLoc :: a, eBin :: BBin } | TB { eLoc :: a, eTer :: BTer } | UB { eLoc :: a, eUn :: BUn }
          | NB { eLoc :: a, eNil :: N }
          | Tup { eLoc :: a, esTup :: [E a] }
+         | Rec { eLoc :: a, esR :: [(Nm a, E a)] }
          | ResVar { eLoc :: a, dfnVar :: DfnVar }
          | RC RurePtr -- compiled regex after normalization
-         | Arr { eLoc :: a, elems :: V.Vector (E a) }
+         | Arr { eLoc :: a, elems :: !(V.Vector (E a)) }
          | Anchor { eLoc :: a, eAnchored :: [E a] }
          | Paren { eLoc :: a, eExpr :: E a }
          | OptionVal { eLoc :: a, eMaybe :: Maybe (E a) }
-         | Cond { eLoc :: a, eIf :: E a, eThen :: E a, eElse :: E a }
-         | In { oop :: E a, ip :: Maybe (E a), mm :: Maybe (E a), istream :: E a }
+         | Cond { eLoc :: a, eIf, eThen, eElse :: E a }
          | RwB { eLoc :: a, eBin :: BBin } | RwT { eLoc :: a, eTer :: BTer }
-         deriving (Functor, Generic)
-
-instance Recursive (E a) where
-instance Corecursive (E a) where
-
-data EF a x = ColumnF a Int
-            | IParseColF a Int | FParseColF a Int | ParseColF a Int
-            | FieldF a Int | LastFieldF a | FieldListF a | AllFieldF a
-            | AllColumnF a | IParseAllColF a | FParseAllColF a | ParseAllColF a
-            | EAppF a x x
-            | GuardedF a x x | ImplicitF a x
-            | LetF a (Nm a, x) x
-            | VarF a (Nm a)
-            | FF (Nm a)
-            | LitF a !L
-            | RegexLitF a BS.ByteString
-            | LamF a (Nm a) x
-            | DfnF a x
-            | BBF a BBin | TBF a BTer | UBF a BUn | NBF a N
-            | TupF a [x]
-            | ResVarF a DfnVar
-            | RCF RurePtr
-            | ArrF a (V.Vector x)
-            | AnchorF a [x]
-            | ParenF a x
-            | OptionValF a (Maybe x)
-            | CondF a x x x
-            | InF x (Maybe x) (Maybe x) x
-            | RwBF a BBin | RwTF a BTer
-            deriving (Generic, Functor)
-
-type instance Base (E a) = (EF a)
+         deriving (Functor)
 
 instance Pretty N where
     pretty Ix="⍳"; pretty Nf="nf"; pretty None="None"; pretty Fp="fp"; pretty MZ="⍬"
@@ -300,12 +273,14 @@ instance PS (E a) where
     ps _ (EApp _ (BB _ op) e) | isJust (mPrec op) = parens (pretty e <+> pretty op)
     ps d (EApp _ (UB _ (At i)) e)     = parensp (d>0) (ps 1 e) <> "." <> pretty i
     ps d (EApp _ (UB _ (Select i)) e) = parensp (d>0) (ps 1 e) <> "->" <> pretty i
+    ps d (EApp _ (UB _ (SelR n)) e)   = parensp (d>0) (ps 1 e) <> "->" <> pretty n
     ps _ (EApp _ (UB _ IParse) e')    = pretty e' <> ":i"
     ps _ (EApp _ (UB _ FParse) e')    = pretty e' <> ":f"
     ps _ (EApp _ (UB _ Parse) e')     = pretty e' <> ":"
     ps d (EApp _ (EApp _ (EApp _ (TB _ Bookend) e) e') e'')  = parensp (d>3) (ps 4 e <> ",," <> ps 4 e' <+> ps 5 e'')
     ps d (EApp _ (EApp _ (EApp _ (TB _ Fold) e) e') e'')     = parensp (d>5) (ps 6 e <> "|" <> ps 6 e' <+> ps 7 e'')
     ps d (EApp _ (EApp _ (EApp _ (TB _ Scan) e) e') e'')     = parensp (d>5) (ps 6 e <> "^" <> ps 6 e' <+> ps 7 e'')
+    ps d (EApp _ (EApp _ (EApp _ (TB _ ScanList) e) e') e'') = parensp (d>5) (ps 6 e <> "^" <> ps 6 e' <+> ps 7 e'')
     ps d (EApp _ (EApp _ (EApp _ (TB _ ZipW) op) e') e'')    = parensp (d>5) ("," <> ps 6 op <+> ps 7 e' <+> ps 8 e'')
     ps d (EApp _ (EApp _ (EApp _ (TB _ Captures) e) e') e'') = parensp (d>6) (ps 7 e <+> "~*" <+> ps 7 e' <+> ps 8 e'')
     ps d (EApp _ e0 e1) = parensp (d>10) (ps 10 e0 <+> ps 11 e1)
@@ -331,7 +306,8 @@ instance PS (E a) where
     ps _ (RwT _ Scan)      = "scan"
     ps _ (RwB _ Fold1)     = "fold1"
     ps _ (Cond _ e0 e1 e2) = "?" <> pretty e0 <> ";" <+> pretty e1 <> ";" <+> pretty e2
-    ps _ (Tup _ es)        = jacTup es
+    ps _ (Tup _ es)        = j'Tup es
+    ps _ (Rec _ es)        = jrec ((\(n,e) -> pretty n <+> ".=" <+> pretty e)<$>es)
     ps _ (Arr _ es)        = tupledByFunky "," (V.toList $ pretty <$> es)
     ps _ (Anchor _ es)     = "&" <> tupledBy "." (pretty <$> es)
     ps _ (Let _ (n, b) e)  = "let" <+> "val" <+> pretty n <+> ":=" <+> pretty b <+> "in" <+> pretty e <+> "end"
@@ -365,6 +341,7 @@ instance Eq (E a) where
     (==) (UB _ unOp) (UB _ unOp')               = unOp == unOp'
     (==) (NB _ x) (NB _ y)                      = x == y
     (==) (Tup _ es) (Tup _ es')                 = es == es'
+    (==) (Rec _ es) (Rec _ es')                 = es == es'
     (==) (ResVar _ x) (ResVar _ y)              = x == y
     (==) (Dfn _ f) (Dfn _ g)                    = f == g -- we're testing for lexical equivalence
     (==) RC{} _                                 = error "Cannot compare compiled regex!"
@@ -394,7 +371,7 @@ instance Show C where show=show.pretty
 -- decl
 data D a = SetFS T.Text | SetRS T.Text
          | FunDecl (Nm a) [Nm a] (E a)
-         | FlushDecl
+         | FlushDecl | SetH
          | SetAsv | SetUsv | SetCsv
          | SetOFS T.Text | SetORS T.Text
          deriving (Functor)
@@ -404,8 +381,10 @@ instance Pretty (D a) where
     pretty (SetRS rs)       = ":set rs :=" <+> "/" <> pretty rs <> "/;"
     pretty (FunDecl n ns e) = "fn" <+> pretty n <> tupled (pretty <$> ns) <+> ":=" <#> indent 2 (pretty e <> ";")
     pretty FlushDecl        = ":flush;"
+    pretty SetH             = ":set header;"
     pretty SetAsv           = ":set asv;"
     pretty SetUsv           = ":set usv;"
+    pretty SetCsv           = ":set csv;"
     pretty (SetOFS sep)     = ":set ofs :=" <+> "'" <> pretty sep <> "';"
     pretty (SetORS sep)     = ":set ors :=" <+> "'" <> pretty sep <> "';"
 
@@ -419,16 +398,19 @@ instance Show (Program a) where show=show.pretty
 flushD :: Program a -> Bool
 flushD (Program ds _) = any p ds where p FlushDecl = True; p _ = False
 
-data Mode = CSV | AWK (Maybe T.Text) (Maybe T.Text) -- field, record
+awk = AWK Nothing Nothing False
+
+data Mode = CSV | AWK (Maybe T.Text) (Maybe T.Text) Bool -- field, record, include header in record split
 
 getS :: Program a -> Mode
-getS (Program ds _) = foldl' go (AWK Nothing Nothing) ds where
-    go (AWK _ rs) (SetFS bs) = AWK (Just bs) rs
-    go _ SetAsv              = AWK (Just "\\x1f") (Just "\\x1e")
-    go _ SetUsv              = AWK (Just "␞") (Just "␟")
-    go _ SetCsv              = CSV
-    go (AWK fs _) (SetRS bs) = AWK fs (Just bs)
-    go next _                = next
+getS (Program ds _) = foldl' go awk ds where
+    go (AWK _ rs b) (SetFS bs) = AWK (Just bs) rs b
+    go _ SetAsv                = AWK (Just "\\x1f") (Just "\\x1e") False
+    go _ SetUsv                = AWK (Just "␞") (Just "␟") False
+    go _ SetCsv                = CSV
+    go (AWK fs _ b) (SetRS bs) = AWK fs (Just bs) b
+    go (AWK fs rs _) SetH      = AWK fs rs True
+    go next _                  = next
 
 mapExpr :: (E a -> E a) -> Program a -> Program a
 mapExpr f (Program ds e) = Program ds (f e)
