@@ -1,7 +1,4 @@
-{-# LANGUAGE DeriveFunctor #-}
-
-module Nm ( Nm (..)
-          , TyName
+module Nm ( Nm (..), TyName
           , eqName
           ) where
 
@@ -14,7 +11,7 @@ data Nm a = Nm { name   :: T.Text
                , loc    :: a
                } deriving (Functor)
 
--- for testing
+-- test
 eqName :: Nm a -> Nm a -> Bool
 eqName (Nm n _ _) (Nm n' _ _) = n == n'
 
@@ -26,7 +23,4 @@ instance Pretty (Nm a) where
 
 instance Show (Nm a) where show=show.pretty
 
-instance Ord (Nm a) where
-    compare (Nm _ u _) (Nm _ u' _) = compare u u'
-
 type TyName = Nm
